//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* commands.cpp: main level commands                                      *//
//* (C) Joachim Wuttke 1990-, C++ rewrite 2001-                            *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <time.h>

#include <ask_simple_value.h>
#include <readln.h>

#include "../config.h"

#include "mystd.h"
#include "olf.h"
#include "mem.h"
#include "edif.h"
#include "func.h" // for initialization
#include "dualplot.h"
#include "opr.h"
#include "plot.h"
#include "file_in.h"
#include "file_out.h"
#include "rssm.h"
#include "manip.h"
#include "curve.h"
#include "special.h"
#include "commands.h"


bool fridaCommand( string cmd )
{
    uint i;

    // Extract overwrite marker:
    NOlm::overwrite = 0;
    if( cmd=="=" )
        throw string( "overwrite marker = must immediately"
                      " follow a command" );
    if ( cmd!="" && cmd[cmd.size()-1]=='=' ) {
        NOlm::overwrite = 1;
        cmd.erase(cmd.size()-1,1);
    }

    if (cmd == "h" || cmd == "?") {
        cout <<
            "This is Frida, version " VERSION "\n"
            "Plain commands:\n"
            "  a    add to plot\n"
            "  p    plot\n"
            "  t    show execution time (for development)\n"
            "  x    execute shell command\n"
            "  quit terminate Frida session\n"
            "Command groups (type initial letter for help):\n"
            "  c    curves and fits\n"
            "  d    directory of online files\n"
            "  e    edit coordinate names and more\n"
            "  g    graphics setup\n"
            "  f    file operations\n"
            "  m    manipulations (reorganisation of data)\n"
            "  o    operations (arithmetic operations on data)\n"
            "  r    raw data read in (currently only for SPHERES)\n"
            "  _    special / temporary commands\n"
            "More help:\n"
            "  \\h   help on meta commands and macros\n"
            "  hm   command modifiers\n"
            "  hf   math operators and functions\n"
            ;

    } else if (cmd == "hm") {
        cout <<
            "Command modifiers:\n"
            "  <cmd>= overwrite the file instead of creating a new one\n"
            ;

    } else if (cmd == "hf") {
        cout << NFunctions::list();

    } else if (cmd == "c") {
        cout << "Curves:\n"
            "  cc   create\n"
            "  cca  create artificial curve (no data reference)\n"
            "Basic operations:\n"
            "  cm   modify function\n"
            "  cd   set data file to fit\n"
            "  cv   set file to convolute with (cv-: no convolution)\n"
            "  cr   set fit range restriction\n"
            "  ci   extract parameter, calculate integral, ...\n"
            "  cp   print parameters\n"
            "Fit to data:\n"
            "  cf   fit\n"
            "  cfs  fit, allow slow convolution\n"
            "  cx   fix parameters\n"
            "  cu   unfix parameters\n"
            "  cwc  weighing: constant\n"
            "  cwl  weighing: logarithmic\n"
            "  cwv  weighing: reciprocal variance (default)\n"
            "Fit tuning parameters:\n"
            "  cs?  show parameters\n"
            "  csf  set gradient step factor\n"
            "  cse  set sqrt of differentiation step\n"
            "  csc  set number of calls\n"
            "  csp  set print frequency\n"
            "  csm  set fit metric\n"
            "  cst  set maximum duration of fits\n"
            "Tuning parameters for integration:\n"
            "  cn?  show parameters\n"
            "  cnm  set integration mode\n"
            "  cna  set eps abs\n"
            "  cnr  set eps rel\n"
            "Tuning parameters for convolution:\n"
            "  cv?  show parameters\n"
            "  cvm  set integration mode\n"
            "  cva  set eps abs\n"
            "  cvr  set eps rel\n"
            ;

    } else if (cmd == "cc" ) {
        NCurveFile::CreateFitcurve();
    } else if (cmd == "cca" ) {
        NCurveFile::CreateFreecurve();

    } else if (cmd == "cm" ) {
        NCurveFile::ChangeExpr();

    } else if (cmd == "cd" || cmd == "cv" || cmd == "cv-" ) {
        NCurveFile::SetFileReference( cmd.substr(1) );

    } else if (cmd == "cr" ) {
        NCurveFile::ChangeRange();

    } else if (cmd == "ci") {
        NCurveFile::IntegralProperty();
    } else if (cmd == "cp") {
        NCurveFile::ShowPar();

    } else if (cmd == "cf") {
        NCurveFile::Fit( false );
    } else if (cmd == "cfs") {
        NCurveFile::Fit( true );
    } else if (cmd == "cx") {
        NCurveFile::SetFixed( true );
    } else if (cmd == "cu") {
        NCurveFile::SetFixed( false );
    } else if (cmd.substr(0,2) == "cw") {
        NCurveFile::SetProperties( cmd.substr(1) );

    } else if (cmd == "c_cd") {
        NCurveFile::CumulativeDistribution();

    } else if (cmd.substr(0,2) == "cs") {
        NCurveFile::SetFitTuningPars(cmd.substr(2));

    } else if (cmd.substr(0,2) == "cn") {
        NCurveFile::SetIntTuningPars(cmd.substr(2));

    } else if (cmd.substr(0,2) == "cv") {
        NCurveFile::SetConvTuningPars(cmd.substr(2));


    } else if ( cmd == "d" ) {
        cout << "Show list:\n"
            "  df   on-line files\n"
            "  dz   z-values and spectral ranges\n"
            "  dp   x|y-values\n"
            "  dy   y-values\n"
            "  dr   real parameters\n"
            "  dc   coordinate names and units\n"
            "  dd   doc lines\n"
            ;

    } else if (cmd == "df") {
        NEdif::ShowFiles();
    } else if (cmd == "dz") {
        NEdif::ShowSpectra();
    } else if (cmd == "dp" ) {
        NOperate::Show( "xyd" );
    } else if (cmd == "dy" ) {
        NOperate::Show( "y" );
    } else if (cmd == "dr") {
        NEdif::ShowNumpar();
    } else if (cmd == "dc") {
        NEdif::ShowCoord();
    } else if (cmd == "dd") {
        NEdif::ShowDoc();

    } else if ( cmd == "e" ) {
        cout << "Edit parameters:\n"
            "  ef         file name\n"
            "  ec<coord>  name and unit of <coord>=x,y,z0,..\n"
            "  ed         documentation lines\n"
            ;

    } else if (cmd.substr(0,2) == "ec") {
        NEdif::EditCoord( cmd.substr(2) );
    } else if (cmd == "ed") {
        NEdif::EditDoc();
    } else if (cmd == "ef") {
        NEdif::EditFNam();


    } else if (cmd == "f") {
        cout << "File operations:\n"
            "  fl   load from disk\n"
            "  fs   save to disk\n"
            "  fs!  save to disk, allow overwrite\n"
            "  fsa  save to disk (old a01 format)\n"
            "  fe1  export (spectrum -> tab-separated line)\n"
            "  fe2  export (as tab)\n"
            "  fc   copy\n"
            "  fx   exchange (copy, then delete source)\n"
            "  fdel delete\n"
            "  fm   make\n"
            "  fr   read in\n"
            "  ft(h|v|c)[m][s]  read tab\n"
            "       h horizontal (only y values, one line per spectrum)\n"
            "       v vertical (one x-y pair per line)\n"
            "       c choose vertical columns\n"
            "       m multi-block (blocks begin with z line)\n"
            "       s from script\n"
            ;

    } else if (cmd == "fl") {
        NFileIn::Load();
    } else if (cmd == "fs") {
        NFileOut::Save( "y08" );
    } else if (cmd == "fs!") {
        NFileOut::Save( "y08", true );
    } else if (cmd == "fsa") {
        NFileOut::Save( "a01" );
    } else if (cmd == "fe1") {
        NFileOut::Save( "csv" );
    } else if (cmd == "fe2") {
        NFileOut::Save( "tab" );
    } else if (cmd == "fc") {
        NOlm::OlfCopy();
    } else if (cmd == "fx") {
        NOlm::OlfExchange();
    } else if (cmd == "fdel") {
        NOlm::OlfDel();
    } else if (cmd == "fm") {
        NEdif::MakeGrid();
    } else if (cmd == "fr") {
        NEdif::ReadIn();
    } else if (cmd.substr(0,2) == "ft") {
        NEdif::ReadTab( cmd.substr(2) );


    } else if (cmd == "g") {
        cout <<
            "Plot:\n"
            "  p    plot\n"
            "  pp   plot with new autoranges\n"
            "  a    add to plot\n"
            "  gd   graphic dialog (feed commands to gnuplot)\n"
            "Save for print:\n"
            "  gp   plot to complete .ps file\n"
            "  gf   plot to short .psX file\n"
            "Change window:\n"
            "  gw   list of plot windows\n"
            "  g<n> switch to plot window <n>\n"
            "Set axes:\n"
            "  gx   set x range\n"
            "  gy   set x range\n"
            "  ga   auto-ranges\n"
            "  gxa  x auto-range\n"
            "  gya  y auto-range\n"
            "  gxl  log x (toggle; or use gxl+ gxl-)\n"
            "  gyl  log y\n"
            "  gxf  force x into frame (toggle; or use gxf+ gxf-)\n"
            "  gyf  force x into frame\n"
            "Other setup:\n"
            "  ge   plot error bars (toggle, or ge+ ge-)\n"
            "  gnd  set maximum number of data points\n"
            "  gnc  number of equidistant base points for curve plotting\n"
            "  gcr  allow iterative refinement for curve plotting?\n"
            "Help:\n"
            "  gsym explain plot symbols\n"
            ;

    } else if ( cmd == "gx" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.ask_and_set( "Plot x range" );
    } else if ( cmd == "gy" ) {
        NPloWin::Plots[NPloWin::iPlot]->Y.ask_and_set( "Plot y range" );
    } else if ( cmd == "ga" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.setAuto();
        NPloWin::Plots[NPloWin::iPlot]->Y.setAuto();
    } else if ( cmd == "gxa" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.setAuto();
    } else if ( cmd == "gya" ) {
        NPloWin::Plots[NPloWin::iPlot]->Y.setAuto();
    } else if ( cmd.substr(0,2) == "gx" ||
                cmd.substr(0,2) == "gy" ||
                cmd.substr(0,2) == "ge" ) {
        NPloWin::Plots[NPloWin::iPlot]->setAux( cmd );
    } else if ( cmd=="gnd" ) {
        NPloWin::Plots[NPloWin::iPlot]->maxpoints =
            iask( "Maximum number of data points in plot",
                  NPloWin::Plots[NPloWin::iPlot]->maxpoints );
    } else if ( cmd=="gnc" ) {
        NPloWin::Plots[NPloWin::iPlot]->equipoints =
            iask( "Number of equidistant base points for curve plotting",
                  NPloWin::Plots[NPloWin::iPlot]->equipoints );
    } else if ( cmd=="gcr" ) {
        NPloWin::Plots[NPloWin::iPlot]->refine =
            bask( "Allow iterative refinement for curve plotting",
                  NPloWin::Plots[NPloWin::iPlot]->refine );
    } else if (cmd == "gp" || cmd == "gf" ) {
        string ps_outdir = NRead::wofmac( "\\psdir" );
        string ps_head = NRead::wofmac( "\\pshead" );
        string ps_dict;
        if ( cmd=="gp" )
            ps_dict = NRead::wofmac( "\\psdict" );
        else
            ps_dict = "";
        NPloWin::Plots[NPloWin::iPlot]->writePostscript(
            ps_outdir, ps_head, ps_dict );
    } else if ( cmd == "gw" ) {
        for( uint i=0; i<NPloWin::nPlot; ++i ){
            printf( " %c %1i %s\n", i==NPloWin::iPlot ? '*' : ' ', i,
                    NPloWin::Plots[i]->info().c_str() );
        }
    } else if ( sscanf( cmd.c_str(), "g%i", &i )==1 ) {
        if( i<NPloWin::nPlot )
            NPloWin::iPlot = i;
        else if ( i==NPloWin::nPlot ){
            NPloWin::iPlot = NPloWin::nPlot;
            NPloWin::Plots.push_back(
                new CPlot( NPloWin::nPlot++, false, false ) );
        } else
            throw string( "invalid graphic window number" );
    } else if (cmd == "gd") { // graph dialog
        string cmd;
        cout << "entering mygnuplot - to leave, type q\n";
        while(1) {
            cmd = sask("mygnuplot> ");
            if (cmd.substr(0,1)=="q")
                break;
            NPloWin::Plots[NPloWin::iPlot]->gp_write(cmd);
        }
    } else if (cmd == "gsym") {
        system( ("gv " + NRead::wofmac( "\\psgsym" ) + "&").c_str() );
    } else if (cmd == "p") {
        NPlot::Plot( NPloWin::Plots[NPloWin::iPlot], 0);
    } else if (cmd == "pp") {
        NPloWin::Plots[NPloWin::iPlot]->X.setAuto();
        NPloWin::Plots[NPloWin::iPlot]->Y.setAuto();
        NPlot::Plot( NPloWin::Plots[NPloWin::iPlot], 0 );
    } else if (cmd == "a") {
        NPlot::Plot( NPloWin::Plots[NPloWin::iPlot], 1 );


    } else if (cmd == "m") {
        cout << "Meta:\n"
            "  m/    protect spectra from forthcoming operations\n"
            "  m/-   lift protections\n"
            "Generic manipulations (expression driven):\n"
            "  md    generic delete\n"
            "  mr    generic retain\n"
            "Per point:\n"
            "  mpd   delete\n"
            "  mpr   retain\n"
            "  mpa   average (=bin), list controlled\n"
            "  mpaf  average, reduce number of points by fixed factor\n"
            "  mpaea average according to error bound (absolute)\n"
            "  mpaer average according to error bound (relative)\n"
            "  mpo   sort\n"
            "  mpq   average when x equal\n"
            "  mpsym symmetrize: map -x -> x, then average\n"
            "Per spectrum:\n"
            "  msd   delete\n"
            "  msr   retain\n"
            "  msa   average\n"
            "  msb   break into files\n"
            "  msj   join\n"
            "  ms*   spawn spectra\n"
            "  msx   z_max <-> x\n"
            "  mso   sort by expression\n"
            "  msoz  sort by z-values\n"
            "  mzx   exchange / rotate z coordinates\n"
            "  mz-   delete z coordinate\n"
            "Per file:\n"
            "  mfj   merge (\"join\")\n"
            "  mfj+  merge, adding z=number-of-input-file\n"
            "  mfjp  merge pointwise\n"
            "  mr-   remove real par\n"
            "Histogram operations:\n"
            "  mhm   bin data AUSSER BETRIEB\n" // TODO
            ;

    } else if (cmd == "m/") {
        NManip::ProtectSpecs( true );
    } else if (cmd == "m/-") {
        NManip::ProtectSpecs( false );
    } else if (cmd == "md") {
        NOperate::Select(1);
    } else if (cmd == "mr") {
        NOperate::Select(0);
    } else if (cmd == "mpd") {
        NManip::PtsSelect( true );
    } else if (cmd == "mpr") {
        NManip::PtsSelect( false );
    } else if (cmd == "mpa") {
        NManip::PtsAvge();
    } else if (cmd == "mpaf") {
        NManip::PtsAvgeFixed();
    } else if (cmd.substr(0,4) == "mpae") {
        NManip::PtsAvgeByErr( cmd.substr(4) );
    } else if (cmd == "mpo") {
        NManip::PtsSort();
    } else if (cmd == "mpq") {
        NManip::PtsAvgeEq();
    } else if (cmd == "mpsym") {
        NManip::PtsSymmetrize();
    } else if (cmd == "msd") {
        NManip::ScaSelect( true );
    } else if (cmd == "msr") {
        NManip::ScaSelect( false );
    } else if (cmd == "msa") {
        NManip::ScaAvge();
    } else if (cmd == "msb") {
        NManip::ScaBreak();
    } else if (cmd == "msj") {
        NManip::ScaJoin();
    } else if (cmd == "ms*") {
        NManip::ScaSpawn();
    } else if (cmd == "msx") {
        NManip::ScaExch();
    } else if (cmd == "mzx") {
        NManip::ZExchange();
    } else if (cmd == "mz-") {
        NEdif::ShowCoordZ();
        NManip::ZDelete();
    } else if (cmd == "mso") {
        NManip::ScaSortByExpr();
    } else if (cmd == "msoz") {
        NManip::ScaSortByZ();
    } else if (cmd == "mfjp") {
        NManip::FilMergePointwise();
    } else if (cmd.substr(0,3) == "mfj") {
        NManip::FilMerge( cmd.substr(3) );
    } else if (cmd == "mr-") {
        NEdif::DelNumpar();
//            } else if (cmd == "mhm") {
//                NManip::HistoMake();


    } else if (cmd == "o") {
        cout << "Operate:\n"
            "  ox       on x\n"
            "  oy       on y\n"
            "  oz0,..   on z0,..\n"
            "  oz+      add new z value\n"
            "  or0,..   on r0,..\n"
            "  or+      add new real par\n"
            "  op0,..   on p0,.. (fit parameter)\n"
            "  op*      on all fit parameters\n"
            "  ofi      functional integrate\n"
            "  ofd      functional derivate\n"
            "  ofac     functional autocorrelate\n"
            "  oi       rank-reducing (""integral"") operations\n"
            "  oio      ditto with oldstyle menu\n"
            "  oixy     choose two columns, save them as x and y\n"
            ;

    } else if (cmd == "op") {
        throw "use op<n> to modify one parameter, or op* to modify all";
    } else if (cmd == "op*") {
        NCurveFile::SetAllPars();
    } else if (cmd[0]=='o' && strchr("xydzpr", cmd[1])) {
        NOperate::Pointwise( cmd.substr(1) );
    } else if (cmd.substr(0,2) == "of") {
        NOperate::Functional( cmd.substr(2) );
    } else if (cmd == "oi") {
        NOperate::Integral();
    } else if (cmd == "oixy") {
        NOperate::IntXY();
    } else if (cmd == "oio") {
        NIntOld::Opr();


    } else if (cmd == "r") {
        cout << "read raw data from SPHERES:\n"
            "  ry   load energy spectra\n"
            "  ry2  load also spectra for open state\n"
            "  ry4  load also chopper histograms\n"
            "  ry8  load spectra for two directions separately\n"
            "  ryd  load only spectra for two directions\n"
            "  ryo  load only spectra for open state\n"
            "  rs   load series of energy spectra\n"
            "  rs2  load also spectra for open state\n"
            ;

    } else if (cmd == "ry" ) {
        NRSSM::ReadSpec( 0 );
    } else if (cmd == "ry2") {
        NRSSM::ReadSpec( 1 );
    } else if (cmd == "ry4") {
        NRSSM::ReadSpec( 3 );
    } else if (cmd == "ry8") {
        NRSSM::ReadSpec( 7 );
    } else if (cmd == "ryd") {
        NRSSM::ReadSpec( 12 );
    } else if (cmd == "ryo") {
        NRSSM::ReadSpec( 9 );
    } else if (cmd == "rs") {
        NRSSM::ReadSeries( 0 );
    } else if (cmd == "rs2") {
        NRSSM::ReadSeries( 1 );


    } else if (cmd == "x") {
        string sh_cmd;
        sh_cmd = sask("shell command");		
        mystd::system( sh_cmd );


    } else if (cmd == "_") {
        cout << "Special programs:\n"
            "  _evv     eigenvalues and eigenvectors\n"
            "  _cc      correlations (y vs y')\n"
            "  _cd      differentiate correlations\n"
            "  _fc      cosine Fourier transform\n"
            "  _t       test (may change at any moment)\n"
//                    "  _ps      power spectrum [disabled]\n"
            ;


    } else if (cmd == "_evv") {
        NSpecial::Eigen();
    } else if (cmd == "_cc") {
        NSpecial::CalcCorrel();
    } else if (cmd == "_cd") {
        NSpecial::CalcCorrelDiff();

    } else if (cmd == "_fc") {
        NSpecial::FourierCosine();

    } else if (cmd == "_t") {
        NSpecial::Test();

    } else if ( cmd == "qui" || cmd=="quit" ) {
        exit(0);

    } else {
        return false; // command not found
    }

    return true;
}
