//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* var.h: variable type and number                                        *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.mesen-und-deuten.de/frida                                   *//
//**************************************************************************//

//! Variable type (and number, if applicable).

class CVariable {
    // Variables may be unnumbered (x,y,i,j,k,c,...) or numbered (z0,p0,r0)
 public:
    enum TTyp {
        _NOREF=0,
        _X, _Y, _Z, _R, // data
        _DY,            // associated error
        _C, _CP, _CQ,   // curve, curve parameters, curve quality
        _K, _J, _I,     // indices for file, spectrum, point   
        _NJ, _NI        // number of spectra in file, points in spectrum
    };
    TTyp typ;
    uint num;

    CVariable( TTyp _typ=_NOREF, uint _num=0 ) : typ(_typ), num(_num) {};
    CVariable( const string s );

    bool is_errvar() const {
        return typ==_DY; }
    bool pointwise() const { 
        return typ==_X || typ==_Y || typ==_DY || typ==_I; };
    bool numbered() const {
        return typ==_CP || typ==_CQ || typ==_Z; };
    // The following are not needed at present:
    //bool named() const {
    //    return typ==_X || typ==_Y || typ==_Z || typ==_R || typ==_CP; }
    //bool in_data() const {
    //    return typ==_X || typ==_Y || typ==_DY || typ==_Z || typ==_R ||
    //        typ==_K || typ==_J || typ==_I || typ==_NJ ||_NI; }
    //bool in_curve() const {
    //    return typ==_X || typ==_Y || typ==_Z || typ==_R ||
    //        typ==_C || typ==_CP || typ==_CQ ||
    //        typ==_K || typ==_J || typ==_NJ; }

    CVariable errvar() const;

    string var_info() const;

 private:
    int requestNum( string s, int pos );
};
